package org.zkoss.less;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.zkoss.io.Files;
import org.zkoss.maven.ZKLessEngine;
import org.zkoss.util.resource.Locator;
import org.zkoss.web.servlet.dsp.Interpretation;
import org.zkoss.web.servlet.dsp.Interpreter;
import org.zkoss.web.servlet.dsp.ServletDspContext;
import org.zkoss.web.util.resource.ClassWebResource;
import org.zkoss.xel.taglib.Taglibs;

import com.asual.lesscss.LessException;
/**
 * This servlet is to compile less into css.dsp file at runtime, used for development phase only.
 * <p>Sample usage:
 * <ol>
 * <li>Add servlet in web.xml</li>
 * <pre>&lt;servlet&gt;
 *     &lt;servlet-name&gt;zkLess&lt;/servlet-name&gt;
 *     &lt;servlet-class&gt;org.zkoss.less.ZKLessServlet&lt;/servlet-class&gt;
 *     &lt;init-param&gt;
 *         &lt;param-name&gt;org.zkoss.less.LessResource&lt;/param-name&gt;
 *         &lt;param-value&gt;/less&lt;/param-value&gt;
 *     &lt;/init-param&gt;
 *     &lt;init-param&gt;
 *         &lt;param-name&gt;org.zkoss.less.OutputFormat&lt;/param-name&gt;
 *         &lt;param-value&gt;.css.dsp&lt;/param-value&gt;
 *     &lt;/init-param&gt;
 *     &lt;init-param&gt;
 *         &lt;param-name&gt;org.zkoss.less.CompressOutput&lt;/param-name&gt;
 *         &lt;param-value&gt;true&lt;/param-value&gt;
 *     &lt;/init-param&gt;
 *     &lt;load-on-startup&gt;1&lt;/load-on-startup&gt;
 * &lt;/servlet&gt;
 * &lt;servlet-mapping&gt;
 *     &lt;servlet-name&gt;zkLess&lt;/servlet-name&gt;
 *     &lt;url-pattern&gt;/less/*&lt;/url-pattern&gt;
 * &lt;/servlet-mapping&gt;</pre>
 * <li>Load less file in zul page</li>
 * <pre>&lt;?link rel="stylesheet" type="text/css" href="path/to/sample.less"?&gt;</pre>
 * </ol>
 * 
 * @author Vincent
 */
public class ZKLessServlet extends HttpServlet {

	private static Logger log = LoggerFactory.getLogger(ZKLessServlet.class);
	private static final long serialVersionUID = 20130826152933L;
	protected int maxAge = 31556926;
	protected long milliseconds = 1000L;

	private static final String LESS_RESOURCE = "org.zkoss.less.LessResource";
	private static final String FORMAT = "org.zkoss.less.OutputFormat";
	private static final String COMPRESS = "org.zkoss.less.CompressOutput";
	private static final String EXT_LESS = ".less";
	private static String OUTPUT_FORMAT = ".css.dsp";
	private static boolean compress = true;
	private static ServletContext sc;
	private static String lessResource;
	private static Map<String, Long> fileModified = new HashMap<String, Long>();
	private ZKLessEngine engine = new ZKLessEngine();

	public void init(ServletConfig config) throws ServletException {
		lessResource = config.getInitParameter(LESS_RESOURCE);
		String param = config.getInitParameter(COMPRESS);
		if (param != null)
			compress = "true".equalsIgnoreCase(param);

		param = config.getInitParameter(FORMAT);
		if (param != null)
			OUTPUT_FORMAT = param;

		sc = config.getServletContext();
		initLessResource();
	}

	public void doGet(HttpServletRequest request, HttpServletResponse response)
			throws ServletException {
		try {
			String reqFilePath = request.getPathInfo();
			String reqRealPath = sc.getRealPath(lessResource + "/" + reqFilePath);
			File lessSource = null;
			File lessOutput = null;
			if (reqFilePath.endsWith(EXT_LESS)) {
				lessSource = new File(reqRealPath);
				lessOutput = new File(reqRealPath.replace(EXT_LESS, OUTPUT_FORMAT));
				Long lessFileLastModified = fileModified.get(reqRealPath);
				// Less file updated
				if (lessFileLastModified == null || lessFileLastModified < lessSource.lastModified()) {
					fileModified
							.put(reqRealPath, lessSource.lastModified());
					compileLess(reqRealPath);
				}
				FileInputStream input = new FileInputStream(lessOutput);
				final Interpretation cnt = 
						parse(input, Interpreter.getContentType(lessOutput.getName()));
				cnt.interpret(
						new ServletDspContext(sc, request, response, null));
			}
		} catch (Exception e) {
			throw new ServletException(e.getMessage(), e);
		}
	}

	protected void doPost(HttpServletRequest req, HttpServletResponse resp)
			throws ServletException, IOException {
		doGet(req, resp);
	}

	private void initLessResource() {
		@SuppressWarnings("unchecked")
		Iterator<String> it = sc.getResourcePaths(lessResource).iterator();
		while (it.hasNext()) {
			String filePath = it.next();
			if (filePath.endsWith(EXT_LESS))
				compileLess(sc.getRealPath(filePath));
		}
	}

	private void compileLess(String lessResourcePath) {
		File input = new File(lessResourcePath);
		File output = new File(
				lessResourcePath.replace(EXT_LESS, OUTPUT_FORMAT));
		fileModified.put(lessResourcePath, input.lastModified());
		log.info("Compiling... " + lessResourcePath);
		try {
			engine.compile(input, output, compress);
		} catch (LessException e) {
			log.error(e.getMessage());
		} catch (IOException e) {
			log.error(e.getMessage());
		}
	}

	private Interpretation parse(InputStream is, String ctype) throws Exception {
		if (is == null)
			return null;

		final String content =
			Files.readAll(new InputStreamReader(is, "UTF-8"))
			.toString();
		final Locator locator = new Locator() {
			public String getDirectory() {
				return null;
			}
			public URL getResource(String name) {
				URL url = null;
				if (name.indexOf("://") < 0) {
					try {
						url = sc.getResource(name);
					} catch (java.net.MalformedURLException ex) {
					}
				}
				return url != null ? url: Taglibs.getDefaultURL(name);
			}
			public InputStream getResourceAsStream(String name) {
				InputStream is = sc.getResourceAsStream(name);
				return is != null ? is:
					ClassWebResource.getClassResourceAsStream(name);
			}
		};
		return new Interpreter().parse(content, ctype, null, locator);
	}
}
